/**
 * @Description: react-redux 封装
 * @Author: lzc
 * @Date: 2019/12/12 17:42:32
 */

import {useSelector, useDispatch} from 'react-redux';

// 所有模块action集合
const ACTIONS = {};
// 所有模块getter集合
const GETTERS = {};

/**
 * 判断是否是对象类型
 * @param v
 * @returns {boolean}
 */
const isObject = v => Object.prototype.toString.call(v) === '[object Object]';

/**
 * 生成每个模块对应的reducer
 * @param namespace {String} 模块名
 * @param state {Object} 对应模块数据
 * @param reducers {Object} 模块reducer
 * @returns {function(*=, *): *} 返回react-redux需要的reducer
 */
const generateReducer = ({ namespace, state, reducers }) => {
  if (isObject(reducers)) {
    return (initState = state, action) => {
      let result = initState;
      const reducersArr = Object.keys(reducers);
      for (let i = 0; i < reducersArr.length; i++) {
        if (typeof reducers[reducersArr[i]] !== 'function') {
          throw new Error('store模块reducers对象属性的值必须为函数类型');
        }
        if (action.type === `${namespace}/${reducersArr[i]}`) {
          const data = reducers[reducersArr[i]](initState, action.payload);
          if (!isObject(data)) {
            throw new Error('store模块reducers对象里的函数必须有返回值，并且为一个对象类型');
          }
          result = { ...initState, ...data };
          break;
        }
      }
      return result;
    };
  }
};

/**
 * 获取所有模块的actions
 * @param namespace
 * @param actions
 * @param getters
 */
const getAllMethods = ({ namespace, actions, getters }) => {
  if (actions && isObject(actions)) {
    Object.keys(actions).forEach(name => {
      if (typeof actions[name] !== 'function') {
        throw new Error('store模块actions对象属性的值必须为函数类型');
      }
      ACTIONS[`${namespace}/${name}`] = actions[name];
    });
  }
  if (getters && isObject(getters)) {
    Object.keys(getters).forEach(getter => {
      GETTERS[`${namespace}/${getter}`] = getters[getter];
    });
  }
};

/**
 * 生成reducer
 * @param modules
 * @returns {{}}
 */
export default modules => {
  if (!isObject(modules)) {
    throw new Error('reduxStore插件必须传入一个对象参数');
  }
  const reducers = {};
  Object.keys(modules).forEach(name => {
    const module = modules[name];
    if (!module.namespace) {
      throw Error('store 模块必须定义namespace属性作为命名空间');
    }
    getAllMethods(module);
    reducers[module.namespace] = generateReducer(module);
  });
  return reducers;
};

/**
 * react-redux 支持异步处理中间件，参考redux-thunk源码
 * https://github.com/reduxjs/redux-thunk/blob/master/src/index.js
 * @param store
 * @returns {function(*): function(...[*]=)}
 */
export const thunk = ({ dispatch, getState }) => next => action => {
  const actionFn = ACTIONS[action.type];
  if (typeof actionFn === 'function') {
    const put = ({ type, payload, ...others }) => {
      const typeName = type.includes(`${action.namespace}/`) ? type : `${action.namespace}/${type}`;
      dispatch({
        type: typeName,
        payload,
        others
      });
    };
    return actionFn({ put, state: getState(), dispatch }, action.payload, action.others);
  } else if (typeof action === 'function') {
    return action(dispatch, getState);
  }
  return next(action);
};

/**
 * 获取getters的hook
 * @param namespace
 * @param getters
 * @returns {{}}
 */
export const useZendGetters = (namespace, getters = []) => {
  const result = {};
  let rootState = {};
  const moduleState = useSelector(state => {
    rootState = state;
    return state[namespace];
  });
  if (Array.isArray(getters)) {
    getters.forEach(getter => {
      const name = `${namespace}/${getter}`;
      if (typeof GETTERS[name] === 'function') {
        result[getter] = GETTERS[name](moduleState, rootState);
      }
    });
  }
  return result;
};

/**
 * 获取Reducer的hook
 * @param namespace
 * @param reducers
 * @returns {{}}
 */
export const useZendReducers = (namespace, reducers) => {
  const dispatch = useDispatch();
  const result = {};
  if (Array.isArray(reducers)) {
    reducers.forEach(reducer => {
      const name = `${namespace}/${reducer}`;
      result[reducer] = (value, ...ret) => dispatch({ type: name, payload: value, namespace, others: ret });
    });
  }
  return result;
};

/**
 * 获取Action的hook
 * @param namespace
 * @param actions
 * @returns {{}}
 */
export const useZendActions = (namespace, actions) => {
  const dispatch = useDispatch();
  const result = {};
  if (Array.isArray(actions)) {
    actions.forEach(action => {
      const name = `${namespace}/${action}`;
      result[action] = (value, ...ret) => dispatch({ type: name, payload: value, namespace, others: ret });
    });
  }
  return result;
};
